/* This is dht.c --  Version 1.2
 * This code is copyright by
 *	Elmar Bartel 1993,
 *	Institut fuer Informatik, TU Muenchen, Germany  
 *	bartel@informatik.tu-muenchen.de
 * You may use this code as you wish, as long as this
 * comment with the above copyright notice is kept intact
 * and in place.
 */
static char what_dynhash[] =
"@(#)dynhash (c) Elmar Bartel 1993";
static char what2_dynhash[] =
"@(#)modified by k.furukawa, kek, v1.2.6 jul.1996-aug.2005";

#include <stdio.h>
#include <stdlib.h>
#include <memory.h>
#ifndef _OSK
#include <stdlib.h>
#endif /* _OSK */
#include "dhtvalue.h"
#include "dht.h"

#define	New(type)	(type *)fxfAlloc(sizeof(type))
#define	nNew(n,type)	(type *)fxfAlloc((n)*sizeof(type))
#define	Nil(type)	(type *)0

/* The next three values are those you may want to change */
#define DefaultMaxLoadFactor	300		/* in percent */
#define DefaultMinLoadFactor	100		/* in percent */
#define ConvertPrime		37
#define	Ld2PtrPerSegment	7
#define MaxDirLevel		8	/* this defines a maximum count of 
					 * buckets to be stored in the table:
					 * (1<<Ld2PtrPerSegment)^(MaxDirLevel+1)
					 */

/* use the one you like most, or is most appropriate for your buisnes
#define	HashPrime		1048583		
#define	HashPrime		2097169
#define	HashPrime		4194353
#define	HashPrime		8388673
#define	HashPrime		16777331
#define	HashPrime		33554693
#define	HashPrime		33554699
#define	HashPrime		67109321
#define	HashPrime		134218627
#define	HashPrime		268437287
#define	HashPrime		536874497
#define	HashPrime		1073749003
#define	HashPrime		2147498033
*/
#define	HashPrime		8388673


#define	PtrPerSegment		(1<<Ld2PtrPerSegment)
#define	SegmentSize		(PtrPerSegment*sizeof(char *))
#define PtrIndexMask		(PtrPerSegment-1)
#define PtrIndex(n)		((n) & PtrIndexMask)
#define	lSegIndex(l,n)		((n)>>((l+1)*Ld2PtrPerSegment))
#define	mlSegIndex(l,n)		(lSegIndex(l,n) & PtrIndexMask)
#define SegIndex(n)		lSegIndex(0,n)
#define mSegIndex(n)		mlSegIndex(0,n)

#define	BucketIndexMask		PtrIndexMask
#define BucketIndex(h)		PtrIndex(h)
#define	SegmentIndex(h)		mSegIndex(h)

#ifndef LOCAL
#define LOCAL
#endif

typedef unsigned long	uLong;
typedef unsigned char	uChar;
typedef unsigned short	uShort;

typedef struct InternHsElement {
	dhtElement		HsEl;
	struct InternHsElement	*Next;
} InternHsElement;
#define NilInternHsElement	Nil(InternHsElement)
#define NewInternHsElement	New(InternHsElement)
#define FreeInternHsElement(h)	fxfFree(h, sizeof(InternHsElement))

uLong	gSegCnt=0;
typedef InternHsElement	*Segment;
#define NilSegment	Nil(Segment)
#define NewSegment	nNew(PtrPerSegment, Segment)
#define FreeSegment(s)	fxfFree(s, PtrPerSegment*sizeof(Segment))


uLong	gDirCnt=0;
typedef Segment		*Directory;
#define NilDirectory	Nil(Directory)
#define NewDirectory	nNew(PtrPerSegment, Directory)
#define FreeDirectory(d) fxfFree(d, PtrPerSegment*sizeof(Directory))


typedef struct {
	uLong		(*Hash)(dhtValue);
	int		(*Equal)(dhtValue, dhtValue);
	dhtValue	(*DupKey)(dhtValue);
	dhtValue	(*DupData)(dhtValue);
	void		(*FreeKey)(dhtValue);
	void		(*FreeData)(dhtValue);
	void		(*DumpData)(dhtValue,FILE *);
	void		(*DumpKey)(dhtValue,FILE *);
} Procedures;

typedef struct {
	Directory	*Dir;
	uShort		Used;		/* fast locating of last */
	uShort		This;		/* its used for stepping through */
} DirectoryTable;

typedef struct dht {
	uLong		p;		/* Next bucket to split */
	uLong		maxp;		/* Upper bound on p during this expansion */ 
	uLong		KeyCount;	/* number keys stored in table */
	uLong		CurrentSize;	/* Current length of table */
	uShort		MinLoadFactor;	/* Lower bound on the load factor */
	uShort		MaxLoadFactor;	/* Upper bound on the load factor */
	DirectoryTable	DirTab[MaxDirLevel];/* The directory table */
	int		DirLevel;
	InternHsElement	*ThisHe;	/* The next Hash-Element to deliver */
	uLong		ThisBucket;	/* The next Bucket to deliver */
	dhtValuePolicy	KeyPolicy;	/* Whether we copy keys */
	dhtValuePolicy	DtaPolicy;	/* Whether we copy data */
	Procedures	procs;
} dht;
#define NewHashTable		(HashTable *)fxfAlloc(sizeof(dht))
#define FreeHashTable(h)	fxfFree(h, sizeof(dht))
#define ActualLoadFactor(h)	((h)->KeyCount*100/(h)->CurrentSize)

unsigned long dhtActualLoad(dht *h) {
	return ActualLoadFactor(h);
}
unsigned long dhtKeyCount(dht *h) {
	return h->KeyCount;
}

dht *dhtCreate(dhtValueType KeyType, dhtValuePolicy KeyPolicy,
			    dhtValueType DtaType, dhtValuePolicy DataPolicy) {
	int		i, SegCnt;
	Directory	*dir;
	dht		*ht;

	SegCnt= 1;	/* this must be smaller thant PtrPerSegment */
	if ((ht=NewHashTable) == NilHashTable) {
	    fprintf(stderr, "dhtCreate: No memory\n");
	    return NilHashTable;
	}
	ht->DirLevel= 0;
	dir= ht->DirTab[0].Dir= NewDirectory;
	if (ht->DirTab[0].Dir == NilDirectory) {
	    fprintf(stderr,
		"dhtCreate: No memory for Directory Segment\n");
	    FreeHashTable(ht);
	    return NilHashTable;
	}
	memset(dir, 0, SegmentSize);
	ht->DirTab[0].Used= SegCnt;
	for (i=0; i<SegCnt; i++) {
	    if ((dir[i]=NewSegment) == NilSegment) {
		fprintf(stderr,
		    "dhtCreate: No memory for Bucket Segment\n");
		SegCnt=i;
		goto Failed;
	    }
	    memset(dir[i], 0, SegmentSize);
	}

	ht->p=
	ht->KeyCount=	    0;
	ht->maxp=	   
	ht->CurrentSize=    SegCnt*PtrPerSegment;
	ht->MinLoadFactor=  DefaultMinLoadFactor;
	ht->MaxLoadFactor=  DefaultMaxLoadFactor;
	ht->KeyPolicy= 	    KeyPolicy;
	ht->DtaPolicy= 	    DataPolicy;

	ht->procs.Hash=     dhtProcedures[KeyType]->Hash;
	ht->procs.Equal=    dhtProcedures[KeyType]->Equal;
	ht->procs.DumpData= dhtProcedures[DtaType]->Dump;
	ht->procs.DumpKey=  dhtProcedures[KeyType]->Dump;

	if (KeyPolicy==dhtNoCopy) {
		ht->procs.DupKey= dhtProcedures[dhtSimpleValue]->Dup;
		ht->procs.FreeKey= dhtProcedures[dhtSimpleValue]->Free;
	}
	else if (KeyPolicy==dhtCopy) {
		ht->procs.DupKey= dhtProcedures[KeyType]->Dup;
		ht->procs.FreeKey= dhtProcedures[KeyType]->Free;
	}
	else {
		fprintf(stderr, "Sorry, unknown KeyPolicy: %d\n", KeyPolicy);
		goto Failed;
	}
	if (DataPolicy==dhtNoCopy) {
		ht->procs.DupData= dhtProcedures[dhtSimpleValue]->Dup;
		ht->procs.FreeData= dhtProcedures[dhtSimpleValue]->Free;
	}
	else if (DataPolicy==dhtCopy) {
		ht->procs.DupData= dhtProcedures[DtaType]->Dup;
		ht->procs.FreeData= dhtProcedures[DtaType]->Free;
	}
	else {
		fprintf(stderr, "Sorry, unknown DataPolicy: %d\n", DataPolicy);
		goto Failed;
	}
	return ht;
Failed:
	for (i=0; i<SegCnt; i++) {
		FreeSegment(ht->DirTab[0].Dir[i]);
		ht->DirTab[0].Dir[i]= NilSegment;
	}
	FreeDirectory(ht->DirTab[0].Dir);
	FreeHashTable(ht);
	return NilHashTable;
}

/* The invariant is
	DirTab[lv].Dir[DirTab[lv].This] = DirTab[lv-1].Dir
 */
LOCAL Directory *SetFirstDirectory(HashTable *ht) {
    int			level= ht->DirLevel;
    DirectoryTable	*dt= &ht->DirTab[level];
    Directory		*dir= dt->Dir;

    dt->This= 0;
    while (level > 0) {
	level--;
	dt--;
	dt->This= 0;
	dt->Dir= dir= (Directory *)dir[0];
    }
    return dir;
}

LOCAL Directory *StepDirectory(HashTable *ht) {
    DirectoryTable *dt= &ht->DirTab[1];
    Directory *dir;
    int level= 1;
    while (level <= ht->DirLevel) {
    	int d= ++dt->This;
    	if (d < PtrPerSegment && dt->Dir[d])
	    break;
	else {
	    dt->This= 0;
	    level++;
	    dt++;
	}
    }
    if (level > ht->DirLevel) {
    	/* to start from the beginning the next time */
    	ht->DirTab[ht->DirLevel].This= 0;
    	return Nil(Directory);
    }
    dir= (Directory *)dt->Dir[dt->This];
    while (level > 1) {
	dt--;
    	level--;
	dt->Dir= dir;
	dir= (Directory *)dir[0];
    }
    dt--;
    /* we are pointing to ht->DirTab[0] */
    if (dt != &ht->DirTab[0]) {
    	fprintf(stderr, "internal error\n");
	exit(0);
    }
    /* this is for Stepping through the various segements */
    dt->This= 0;
    dt->Dir= dir;
    return dir;
}

LOCAL Directory *StepAndDeleteDirectory(HashTable *ht) {
    DirectoryTable *dt= &ht->DirTab[1];
    Directory *dir;
    int level= 1;
    while (level <= ht->DirLevel) {
    	int d= dt->This;
	FreeSegment(dt->Dir[d]);
	dt->Dir[d]= NilSegment;
	d++;
    	if (d < PtrPerSegment && dt->Dir[d]) {
	    dt->This++;
	    break;
	}
	else {
	    dt->This= 0;
	    level++;
	    dt++;
	}
    }
    if (level > ht->DirLevel) {
    	FreeSegment(ht->DirTab[ht->DirLevel].Dir);
    	ht->DirTab[ht->DirLevel].Dir= Nil(Directory);
    	return Nil(Directory);
    }
    dir= (Directory *)dt->Dir[dt->This];
    while (level > 1) {
	dt--;
    	level--;
	dt->Dir= dir;
	dir= (Directory *)dir[0];
    }
    return dir;
}

Directory *SetDirectory(HashTable *ht, uLong newp) {
    static char *myname= "SetDirectory";
    int level, d;
    Directory *dir;
    level= ht->DirLevel;
    if ((d=lSegIndex(level, newp)) >= PtrPerSegment) {
	/* we have to bump the DirLevel */
	DirectoryTable *dt; 
	if (level+1 >= MaxDirLevel) {
	    fprintf(stderr, "%s: exceeding MaxDirLevel (=%d)\n",
	    	myname, MaxDirLevel);
	    return Nil(Directory);
	}
	level++;
	dt= &ht->DirTab[level];
	if ((dt->Dir=NewDirectory) == NilDirectory) {
	    fprintf(stderr, "%s: no mem for %d-level directory\n",
	    	myname, level);
	    return Nil(Directory);
	}
	memset(dt->Dir, '\0', SegmentSize);
	dt->Dir[0]= (Segment *)dt[-1].Dir;  /* the old top directory */
	dt->Used= 1;
	level= ++ht->DirLevel;
	d=lSegIndex(level, newp);
    }
    ht->DirTab[level].This= d;
    dir= ht->DirTab[level].Dir;
    while (level > 0) {
    	if (((Directory *)dir[d]) == NilDirectory) {
	    if ((dir[d]=NewSegment) == NilSegment) {
		fprintf(stderr, "%s: no mem for %d-level directory\n",
		    myname, level);
		return Nil(Directory);
	    }
	    memset(dir[d], '\0', SegmentSize);
	    ht->DirTab[level].Used++;
	    ht->DirTab[level-1].Used= 0;
	}
    	level--;
	ht->DirTab[level].Dir= dir= (Directory *)dir[d];
	d= ht->DirTab[level].This= mlSegIndex(level, newp);
    }
    return dir;
}

void dhtDestroy(HashTable *ht) {
    Directory *dir= SetFirstDirectory(ht);

    while (dir) {
	Segment *seg;
	int d= 0;
	while (d<PtrPerSegment && (seg=dir[d])) {
	    int i;
	    for (i=0; i<PtrPerSegment; i++) {
		InternHsElement *b= seg[i];
		while (b) {
		    InternHsElement *tmp= b;
		    (ht->procs.FreeKey)(b->HsEl.Key);
		    (ht->procs.FreeData)(b->HsEl.Data);
		    b= b->Next;
		    FreeInternHsElement(tmp);
		}
	    }
	    FreeSegment(seg);
	    dir[d]= NilSegment;
	    d++;
	}
	dir= StepAndDeleteDirectory(ht);
    }
    FreeHashTable(ht);
}

void dhtDumpIndented(int ind, HashTable *ht, FILE *f) {
    static char *myname= "dhtDumpIndented";
    Directory *dir;
    int d, dcnt, hcnt;
    char ib[256];

    if (ind > sizeof(ib)) {
	fprintf(stderr, "%s: Cannot indent for %d chars\n",
	    myname, ind);
	exit(5);
    }
    for (d=0; d<ind; d++)
	ib[d]= ' ';
    ib[ind]= '\0';

    fprintf(f, "%sSimple Values: \n", ib);
    fprintf(f, "%sp (Next bucket to split) = %6u\n", ib, ht->p);
    fprintf(f, "%smaxp (Upper bound on p)  = %6u\n", ib, ht->maxp);
    fprintf(f, "%sKeyCount                 = %6u\n", ib, ht->KeyCount);
    fprintf(f, "%sCurrentSize              = %6u\n", ib, ht->CurrentSize);
    fprintf(f, "%sMinLoadFactor (%%)        = %6u\n", ib, ht->MinLoadFactor);
    fprintf(f, "%sMaxLoadFactor (%%)        = %6u\n", ib, ht->MaxLoadFactor);
    fprintf(f, "%sActual LoadFactor (%%)    = %6d\n", ib, ActualLoadFactor(ht));
    fprintf(f, "%sDirLevel                 = %6u\n", ib, ht->DirLevel);
    dcnt= hcnt=0;
    /*
    for (d=0; d<ht->DirUsed; d++) {
	InternHsElement **seg= ht->Dir[d];
	int i;
	fprintf(f, "%sSegment %d\n", ib, d);
	for (i=0; i<PtrPerSegment; i++) {
	    InternHsElement *b;
	    if ((b=seg[i]) == NilInternHsElement)
		continue;
	    fprintf(f, "%s  Bucket %d:\n", ib, i);
	    while (b) {
	    	fputs(ib, f); fputs("    ", f);
		(ht->procs.DumpKey)(b->HsEl.Key, f);
		fputs("->", f);
		(ht->procs.DumpData)(b->HsEl.Data, f);
		b= b->Next;
		fputc('\n', f);
		hcnt++;
	    }
	}
    }
    */
    dir= SetFirstDirectory(ht);
    while (dir) {
    	int d= 0;
	Segment *seg;
	while (d<PtrPerSegment && (seg=dir[d])) {
	    int i;
	    fprintf(f, "%sSegment %d\n", ib, dcnt);
	    for (i=0; i<PtrPerSegment; i++) {
		InternHsElement *b= seg[i];
		fprintf(f, "%s  Bucket %d:\n", ib, i);
		if ((b=seg[i]) == NilInternHsElement)
		    continue;
		while (b) {
		    fputs(ib, f); fputs("    ", f);
		    (ht->procs.DumpKey)(b->HsEl.Key, f);
		    fputs("->", f);
		    (ht->procs.DumpData)(b->HsEl.Data, f);
		    b= b->Next;
		    fputc('\n', f);
		    hcnt++;
		}
	    }
	    d++;
	    dcnt++;
	}
	dir= StepDirectory(ht);
    }
    fprintf(f, "%s%d records of %d dumped\n\n", ib, hcnt, ht->KeyCount);
}

void dhtDump(HashTable *ht, FILE *f) {
	dhtDumpIndented(0, ht, f);
}

static void StepElement(HashTable *ht) {
	uLong d,i;
	DirectoryTable *dt;
	Directory *dir;

	if (ht->KeyCount == 0 || ht->ThisHe == NilInternHsElement)
		return;
	if (ht->ThisHe=ht->ThisHe->Next)
		return;

	dt= &ht->DirTab[0];
	dir= dt->Dir;
	i= ht->ThisBucket+1;
	d= dt->This;
	do {
	    Segment *seg;
	    while (d<PtrPerSegment && (seg=dir[d])) {
		while (i<PtrPerSegment) {
		    if (seg[i]) {
			dt->This= d;
			ht->ThisHe= seg[i];
			ht->ThisBucket= i;
			return;
		    }
		    else
			i++;
		}
		d++;
		i= 0;
	    }
	    d= 0;
	} while (dir=StepDirectory(ht));
	return;
}

dhtElement *dhtGetFirstElement(HashTable *ht) {
    Directory *dir;

    if (ht->KeyCount == 0)
	    return dhtNilElement;

    dir= SetFirstDirectory(ht);
    while (dir) {
	int d= 0;
	Segment *seg;
	while (d<PtrPerSegment && (seg=dir[d])) {
	    int i;
	    for (i=0; i<PtrPerSegment; i++) {
		if (seg[i]) {
		    ht->ThisHe= seg[i];
		    ht->ThisBucket= i;
		    StepElement(ht);
		    return &seg[i]->HsEl;
		}
	    }
	    d++;
	}
	dir= StepDirectory(ht);
    }
    return dhtNilElement;
}

dhtElement *dhtGetNextElement(HashTable *ht) {
    if (ht->ThisHe) {
    	dhtElement *de= &ht->ThisHe->HsEl;
	StepElement(ht);
	return de;
    }
    else 
	return dhtNilElement;
}

LOCAL uLong DynamicHash(uLong p, uLong maxp, uLong v) {
	uLong h, k= v % HashPrime; 
	if ((h = k%maxp) < p)
		return k % (maxp<<1);
	else
		return h;
}

LOCAL dhtStatus ExpandHashTable(HashTable *ht) {
    static char *myname= "ExpandHashTable";
    /* Need to expand the directory */
    Directory	*odir= SetDirectory(ht, ht->p);
    uLong	olds= SegmentIndex(ht->p);
    uLong	oldb= BucketIndex(ht->p);
    uLong	newp= ht->maxp + ht->p;
    Directory	*ndir= SetDirectory(ht, newp);
    uLong	news= SegmentIndex(newp);
    uLong	newb= BucketIndex(newp);
    int level, d;


#ifdef NOTDEF
    if (news >= ht->DirSize) {
	Segment **d;
	int i, SegCnt;
	SegCnt= ht->DirSize;
	SegCnt+= (SegCnt>>3) + 1;
	d= (Directory *)fxfReAlloc(ht->Dir,
	  ht->DirSize*sizeof(Directory), SegCnt*sizeof(Directory));
	if (d == (Segment **)0) {
	    fprintf(stderr, "%s: No memory\n", myname);
	    return dhtFailedStatus;
	}
	ht->DirSize= SegCnt;
	ht->Dir= d;
    }
    if (news >= ht->DirUsed) {
	if ((ht->Dir[news]=NewSegment) == NilSegment) {
	    fprintf(stderr, "%s: No memory\n", myname);
	    return dhtFailedStatus;
	}
	memset(ht->Dir[news], 0, SegmentSize);
	ht->DirUsed++;
    }
#endif
    if (ndir[news] == NilSegment) {
	if ((ndir[news]=NewSegment) == NilSegment) {
	    /*fprintf(stderr, "%s: No memory\n", myname);*/
	    return dhtFailedStatus;
	}
	memset(ndir[news], 0, SegmentSize);
	ht->DirTab[0].Used++;
    }
    /* update bucket pointers */
    ht->p++;
    if (ht->p == ht->maxp) {
	ht->maxp<<= 1;
	ht->p= 0;
    }
    ht->CurrentSize++;
    /* relocate records */
    {
	InternHsElement **new= &(ndir[news][newb]);
	InternHsElement **old= &(odir[olds][oldb]);
	
	while (*old) {
	    if (DynamicHash(ht->p, ht->maxp,
	     (ht->procs.Hash)((*old)->HsEl.Key)) == newp) {
		*new= *old;
		*old= (*old)->Next;
		new= &(*new)->Next;
		*new= NilInternHsElement;
	    }
	    else
		old= &(*old)->Next;
	}
    }
    return dhtOkStatus;
}

LOCAL void ShrinkHashTable(HashTable *ht) {
    InternHsElement	**old, **new;
    Directory	*ndir, *odir;
    int		od;
    uLong		oldp;

    if (ht->maxp == PtrPerSegment && ht->p == 0)
	    return;
    if (ht->p == 0) {
	    ht->maxp>>= 1;
	    ht->p= ht->maxp;
    }
    ht->p--;
    ndir=SetDirectory(ht, ht->p);
    new= &(ndir[SegmentIndex(ht->p)][BucketIndex(ht->p)]);
    oldp= ht->p + ht->maxp;
    odir=SetDirectory(ht, oldp);
    old= &(odir[od=SegmentIndex(oldp)][BucketIndex(oldp)]);
    if (*old) {
	while (*new)
		new= &(*new)->Next;
	*new= *old;
	*old= NilInternHsElement;
    }
    ht->CurrentSize--;

    if (BucketIndex(oldp) == 0) {
	if (ht->DirLevel == 0) {
	    if (od > 0) {
	    	if (od > PtrPerSegment) {
			fprintf(stderr, "Impossible\n");
#ifndef _OSK
			kill(getpid(), 3);
#else /* _OSK */
			exit(1);
#endif /* _OSK */
		}
		FreeSegment(odir[od]);
		odir[od]= NilSegment;
		ht->DirTab[0].Used--;
	    }
	}
	else {
	    FreeSegment(odir[od]);
	    odir[od]= NilSegment;
	    if (--ht->DirTab[0].Used == 0) {
		int d, level= 1;
		while (level <= ht->DirLevel) {
		    d= --ht->DirTab[level].Used;
		    FreeDirectory(ht->DirTab[level].Dir[d]);
		    ht->DirTab[level].Dir[d]= NilSegment;
		    if (d == 0)
			level++;
		    else
			break;
		}
		if (level == ht->DirLevel && d==1) {
		    /* only one dir in top - so decrease level */
		    int lv= --ht->DirLevel;
		    ht->DirTab[lv].Dir= (Directory *)ht->DirTab[lv+1].Dir[0];
		    FreeSegment(ht->DirTab[lv+1].Dir);
		    ht->DirTab[lv+1].Dir= Nil(Directory);
		}
		while (level >= 0) {
		    if (ht->DirTab[level].Used == 0)
			ht->DirTab[level].Used= PtrPerSegment;
		    level--;
		}
	    }
	}
    }
    return;
}

LOCAL InternHsElement **LookupInternHsElement(HashTable *ht, dhtValue key) {
	uLong h;
	InternHsElement **phe;
	Directory *dir;
	int level= ht->DirLevel;
	h= DynamicHash(ht->p, ht->maxp, (ht->procs.Hash)(key));
	dir= ht->DirTab[level].Dir;
	while (level > 0) {
	    dir= (Directory *)dir[mlSegIndex(level, h)];
	    level--;
	}
	phe= &(dir[SegmentIndex(h)][BucketIndex(h)]);
	while (*phe) {
		if ((ht->procs.Equal)((*phe)->HsEl.Key, key))
			break;
		else
			phe= &((*phe)->Next);
	}
	return phe;
}

void dhtRemoveElement(HashTable *ht, dhtValue key) {
	static char *myname= "dhtRemoveElement";
	uLong h;
	InternHsElement **phe, *he;

	phe= LookupInternHsElement(ht, key);
	if (*phe) {
		/*
		fprintf(stderr,
		  "%s: dumping before removing\n", myname);
		fDumpHashTable(ht, stderr);
		*/
		he= *phe;
		if (ht->ThisHe == he)
			StepElement(ht);
		*phe= he->Next;
		(ht->procs.FreeData)(he->HsEl.Data);
		(ht->procs.FreeKey)(he->HsEl.Key);
		FreeInternHsElement(he);
		ht->KeyCount--;
		if (ActualLoadFactor(ht) < ht->MinLoadFactor) {
		    /*
		    fprintf(stderr,
		      "%s: dumping before shrinking\n", myname);
		    fDumpHashTable(ht, stderr);
		    */
		    ShrinkHashTable(ht);
		    /*
		    fprintf(stderr,
		      "%s: dumping after shrinking\n", myname);
		    fDumpHashTable(ht, stderr);
		    */
		}
		/*
		fprintf(stderr, "%s: dumping after removing\n", myname);
		fDumpHashTable(ht, stderr);
		*/
	}
	return;
}

dhtStatus	dhtDupStatus;

dhtElement *dhtEnterElement(HashTable *ht, dhtValue key, dhtValue data) {
	InternHsElement **phe, *he;
	dhtValue KeyV, DataV;
	dhtStatus stat= dhtOkStatus;

	KeyV= (ht->procs.DupKey)(key);
	if (dhtDupStatus != dhtOkStatus)
		return dhtNilElement;
	DataV= (ht->procs.DupData)(data);
	if (dhtDupStatus != dhtOkStatus) {
		(ht->procs.FreeKey)(KeyV);
		return dhtNilElement;
	}
	phe= LookupInternHsElement(ht, key);
	he= *phe;
	if (!he) {
		if (!(he=NewInternHsElement))
			return dhtNilElement;
		*phe= he;
		he->Next= NilInternHsElement;
		ht->KeyCount++;
	}
	else {
		if (ht->DtaPolicy == dhtCopy)
			(ht->procs.FreeData)(he->HsEl.Data);
		if (ht->KeyPolicy == dhtCopy)
			(ht->procs.FreeKey)(he->HsEl.Key);
	}
	he->HsEl.Key= KeyV;
	he->HsEl.Data= DataV;
	if (ActualLoadFactor(ht) > ht->MaxLoadFactor) {
	    /*
	    fprintf(stderr, "Dumping Hash-Table before expansion\n");
	    fDumpHashTable(ht, stderr);
	    */
	    if ((stat=ExpandHashTable(ht)) != dhtOkStatus)
		    return dhtNilElement;
	    /*
	    fprintf(stderr, "Dumping Hash-Table after expansion\n");
	    fDumpHashTable(ht, stderr);
	    */
	}
	return &he->HsEl;
}

dhtElement *dhtLookupElement(HashTable *ht, dhtValue key) {
	InternHsElement **phe= LookupInternHsElement(ht, key);
	if (*phe)
		return &(*phe)->HsEl;
	else
		return dhtNilElement;
}
