/*
 *	rtbl.h	n.k.	June 14, 1992
 *
 * used as "server.h" in SCLIB-SERVER
 * use "tbl.h" for local error definition (in the middle of this include file)
 *
 * Modifications :
 *   k.f. apr.21.1996  rename sc_arg into sc_scarg since signal.h on hp use it
 *                     many symbols like sc_xxx are used, be careful in future
 *
 * not yet optimised for version 2
**/

/**
 *  definitions : to be revised for each program
 *
 * note: SC_S_EXTFUNC does not need quotations ("")
 * note: if SINGLESERVER is defined in netdef.h, SC_S_SINGLESERVER should 
 *       not be zero
**/
/* general definitions */
#define SC_S_WHAT "@(#)rtbld v1.0a, remote table (rtbllib) manager, N. Kamikubota, June 1992"
#define SC_S_SERVICE "#4811"		/* service name or #port_number */
#define SC_S_SERVHOST "peach"		/* service hostname */
#define SC_S_NM  "rtbld"		/* daemon process name */
#define SC_S_EXTFUNC _tbl		/* name of the user-function */

/* options for both server and client */
/* need compilation/link  of both server and client */
#define SC_S_SINGLESERVER  0
	/* =0 -- folk a child for each request of client */
	/* =1 -- invoke open/close procedure for each process */
	/*       2nd process wait for completion of the 1st one (no timeout) */
	/* =2 -- invoke open/close procedure for each call, each time */

/* server options */
#define SC_S_LINACNET	1
	/* =0 -- accept calls from any workstation */
	/* =1 -- disconnect if client call is from out-of Linac */
#define SC_S_CONNMSG 1			/* =1 recommended */
	/* =0 -- no message of client_host, port #, connection time, etc. */
	/* =1 -- show message at socket connection and disconnection */

/* client options */
#define SC_S_CLIWARNING 1
	/* =0 -- return error code when error in client.c */
	/* =1 -- show warning mesage & exit when error in client.c */

/* additional options (additional to those realized without SCLIB-SERVER
) */
#define SC_S_COMCHECK	1		/* =0 is default */
	/* =0 -- user_func() judge the validity of specified "com" */
	/*       even that one can judge it in client.c */
	/* =1 -- check "com" in client.c, return error immediately */
	/*       version of server.h should be same at both cli/serv sites */

/***************************************************************************/
/****************  DO  NOT  CHANGE  HERE  **********************************/
									/***/
/* version conflict */							/***/
#ifdef SC_L_UARG							/***/
  !! [33m[7m USE SC_LIB.H AFTER V1.5H [37m[0m !! ... produced in server.h
#endif									/***/
									/***/
/* error codes : reserved in client.c */				/***/
#define SC_SE_SCLIB	(-81)	/* error in SCLIB */			/***/
#define SC_SE_NETPROT	(-82)	/* call from out-of the linac */	/***/
#define SC_SE_SCOPEN	(-83)	/* unable to connect to server */	/***/
#define SC_SE_LIBRBUG	(-20)	/* other error ; library bug ? */	/***/

/* error codes : used when SC_S_COMCHECK */
#if ( SC_S_COMCHECK == 1 )
#define SC_SE_COMMAND   (-11)   /* illegal command */
#endif
									/***/
/****************  DO  NOT  CHANGE  HERE  **********************************/
/***************************************************************************/

/**
 *  error codes : user-defined in SC_S_EXTFUNC
**/
#define TBL_E_SCLIB SC_SE_SCLIB		/* (-81) error in SCLIB */
#define TBL_E_NETPROT SC_SE_NETPROT	/* (-82) call from out-of the linac */
#define TBL_E_SCOPEN SC_SE_SCOPEN	/* (-83) unable to connect to server */
#define TBL_E_LIBRBUG SC_SE_LIBRBUG	/* (-20), library bug ? */

/* SC_S_COMMAND defined/undefined dep. on SC_S_COMCHECK */
#if ( SC_S_COMCHECK == 1 )
#define TBL_E_COMMAND SC_SE_COMMAND     /* (-11), illegal command */
#else
#define TBL_E_COMMAND (-11)		/* illegal command */
#endif

/* followings are defined in SC_S_EXTFUNC (_tbl.c) */
#include	"tb2.h"			/* error definitions */

/**
 *  definitions of buffer size for "sc_uarg", "sc_scarg", and "sc_com"
 *
 * note: header size is fixed as 64 bytes; header contains COM and NAME 
 * note: SC_L_UARG is the size of data buffer for data exchange
 *       default size is 64, which corresponds to int[16], float[8], char[64]
 * note: thus packet size for data exchange is (64 + SC_L_UARG) bytes
 *       do net exceed SCLIB buffer, SC_BUFSIZ (default is 1024)
**/

#define SC_L_COM 16		/* length of com[] in sc_scarg, usually 16 */
#define SC_L_NAME 32		/* length of name[] in sc_scarg, usually 32 */
#define SC_L_UARG 64		/* length of union uarg, data buffer size */

/***************************************************************************/
/****************  DO  NOT  CHANGE  HERE  **********************************/
									/***/
union sc_uarg {		/* data buffer */				/***/
	int	iarg;		/* int */				/***/
	int	idim[SC_L_UARG/4];					/***/
	long	larg;		/* long */				/***/
	long	ldim[SC_L_UARG/4];					/***/
	short	sarg;		/* short */				/***/
	short	sdim[SC_L_UARG/2];					/***/
	float	farg;		/* float */				/***/
	float	fdim[SC_L_UARG/8];					/***/
	double	darg;		/* double precision */			/***/
	double	ddim[SC_L_UARG/8];					/***/
	char	carg;		/* char */				/***/
	char	cdim[SC_L_UARG];					/***/
};									/***/
									/***/
#define SC_L_ARG (64+SC_L_UARG)	/* length of strunt sc_scarg */		/***/
#if ( (64-SC_L_COM-SC_L_NAME-13) <= 0 )					/***/
  !! [33m[7m COM/NAME LENGTH IS TOO LONG [37m[0m !! ... produced in server.h
#endif									/***/
									/***/
struct sc_scarg {	/* packet for data exchange, header size = 64 *//***/
	char	com[SC_L_COM+1];   /* command to a device */		/***/
	char	name[SC_L_NAME+1]; /* name of a device to be used */	/***/
	char	ptype;		/* parameter type, 'C', 'F', 'I', 'B', '0'**/
	char	cdummy[64-SC_L_COM-SC_L_NAME-15];			/***/
	long	rtn;		/* return code */			/***/
	long	narg;		/* # of arguments */			/***/
	long	buf;		/* additional buffer */			/***/
	union sc_uarg  param;	/* parameter field */			/***/
};									/***/
									/***/
struct sc_com {		/* struct definition */				/***/
	char	com[SC_L_COM+1];/* command */				/***/
	char	intype;		/* type of input parameters */		/***/
	int	nin;		/* # of input parameters */		/***/
	char	outtype;	/* 'C','I','F','B', or '0' */		/***/
	int	nout;		/* # of output parameters */		/***/
};									/***/
									/***/
/****************  DO  NOT  CHANGE  HERE  **********************************/
/***************************************************************************/


/**
 *  list for commands available in SC_S_EXTFUNC()
 *
 * note: intype/outtype is one of 'C','I','F','B','0'
 *       'C' -- character string with null termination, assume nin/nout=1
 *       'I' -- int or int-dimension
 *       'F' -- float or float-dimesion (nin/out max is 4 at present vesion)
 *       'B' -- binary data, treated as char*, nin/out is # of bytes
 *       '0' -- no data needed, accordingly nin/nout = 0
 * note: SC_L_UARG is the size of data buffer for data exchange
 *       default size is 64, which corresponds to int[16], float[8], char[64]
 *       do not exceed this number when you change sc_s_com[]
**/

#define SC_S_NCOM 4	/* number of commands */
static struct sc_com	sc_s_com[] = {
	{ "READ", '0',0, '0',0 },	/* icom = 0 */
	{ "PARAM", '0',0, 'I',1 }, 	/* icom = 1 */
	{ "INT", 'C',1, 'I',1 },	/* icom = 2 */
	{ "NODE", '0',0, 'C',1 }	/* icom = 3 */
};
