/**
 *  tbllist  - get name list for specified node
 *  n.k. dec.21 1994
 *  tb2list  - get name list for specified node
 *  k.f. oct.1 1996
 *
 * Modifications:
 *  n.k. july26,1995  show version (v1.4)
 *
 * version 2.0 does not support "-n", "-p" options
**/
static char what_tb2list[] =
"@(#)tb2list.c, v2.0.1, n.f, k.f, linac table list, dec.1994-oct.1996";

/*** definitions ***/
#include  <stdio.h>
#include  "tb2.h"

void printnames( tb2_Spec *tb2, int nodeindex, char * node, int firstserial );

/**********************
 *   main of tbllist
**/
main( argc, argv )
int    argc;
char   *argv[];
{
    int    i_tblarg=1, i_nodearg=2;	/* argv[] pointer */
    int    usen=0, usep=0;		/* option flag */
    tb2_Spec *tb2;			/* table specification */
    tb2_ParamEnt(*parament)[];
    int serial, firstserial, lastserial;	/* serial number */
    int nodeindex;			/* index to "NODE" */
    char *node, *prevnode, *nextnode;	/* node names */

/** argument handling **/
    switch( argc ) {
/* #ifdef OPTIONSNP */ /* not supported */
    case 4:
	if( *argv[1] == '-' ) {
	    switch( *(argv[1] + 1) ) {
	    case 'N':      case 'n':            /* option N */
		usen++;
		break;
	    case 'P':      case 'p':            /* option P */
		usep++;
		break;
	    default:
		fprintf( stderr, "%s - invalid option '%s' specified\n\n", 
			argv[0], argv[1] );
	    case 'H':      case 'h':            /* go to help */
		goto USAGE;
	    }
	} else {
	    fprintf( stderr, "%s - specify option first\n\n", argv[0] );
	    goto USAGE;
	}
	i_tblarg=2;i_nodearg=3;
	break;
/* #endif */ /* OPTIONSNP */
    case 3:
	i_tblarg=1;i_nodearg=2;
	break;
    default:
	goto USAGE;

    }
    node = argv[i_nodearg];

/** read table **/
    tbl_read( argv[i_tblarg] );
    if( tb2_errno < 0 )  tbl_errend();
    tb2 = default_tbl;

/** node keyword search **/
    if ( -1 == (nodeindex = tb2_keywordfind( tb2, "NODE" )) )
	if( tb2_errno < 0 )  tbl_errend();

/** name search [pass1] - set n_nodes, i_node/i **/
    firstserial = lastserial = 0;
    prevnode = nextnode = (char *)NULL;
    for ( serial = 1; 
	 (parament = (tb2_ParamEnt(*)[]) tb2_sentry( tb2, serial ));
	 serial++ )  {
	if ( !firstserial )  {
	    if ( !strcasecmp( (*parament)[nodeindex].string, node ) )  {
		firstserial = serial;		/* first matching entry */
	    }  else  {				/* not match yet */
		prevnode = (*parament)[nodeindex].string;
	    }					 /* prevnode candidate */
	}  else  {
	    lastserial = serial - 1;		/* last(maybe) match entry */
	    if ( strcasecmp( (*parament)[nodeindex].string, node ) )  {
		nextnode = (*parament)[nodeindex].string;
		break;
	    }
	}
    }
    if ( !firstserial )
	tbl_errend();
    if ( !lastserial )
	tbl_errend();
#ifdef DEBUG1
    fprintf( stderr, 
	    "firstserial=%d, lastserial=%d, prevnode=%s, nextnode=%s\n",
	    firstserial, lastserial, prevnode, nextnode );
#endif /* DEBUG1 */

/** name search [pass2] - show namelist **/

    if ( !(usen-usep) )  {
	printnames( tb2, nodeindex, node, firstserial );
    }  else if ( usep && prevnode )  {
	printnames( tb2, nodeindex, prevnode, 1 );
    }  else if ( usen && nextnode )  {
	printnames( tb2, nodeindex, nextnode, 1 );
    }

    exit( 0 );

/** usage **/
USAGE:
    fprintf( stderr, "%s %s - get namelist from the table\n\n", 
	    argv[0], TBL_VER );
    fprintf( stderr, "Usage: %s <option> abbr node\n", argv[0] );
    fprintf( stderr, "       abbr -- abbreviation of table filename\n" );
    fprintf( stderr, "       node -- node name to be searched\n" );
    fprintf( stderr, "Option:  -h help message\n" );
#ifdef OPTIONSNP /* not supported */
    fprintf( stderr, "         -n search next node \n" );
    fprintf( stderr, "         -p search previous node \n" );
    fprintf( stderr, "Example: %s -n mgv3 L3-1-2\n", argv[0]);
    fprintf( stderr, "         %s -p loop3 mutsuki\n", argv[0]);
    fprintf( stderr, "         %s dac12 saburo\n", argv[0]);
    fprintf( stderr, "Note: n/p-option with invalid node results in name-list for the initial node\n" );
#else /* OPTIONSNP */
    fprintf( stderr, "Example: %s mgv3 L3-1-2\n", argv[0]);
    fprintf( stderr, "         %s dac12 saburo\n", argv[0]);
#endif /* OPTIONSNP */
    exit( 1 );
}

void printnames( tb2_Spec *tb2, int nodeindex, char * node, int firstserial )
{
    int serial;
    tb2_ParamEnt(*parament)[];

#ifdef DEBUG1
    fprintf( stderr, 
	    "firstserial=%d, node=%s\n",
	    firstserial, node );
#endif /* DEBUG1 */

    for ( serial = firstserial;
	 (parament = (tb2_ParamEnt(*)[]) tb2_sentry( tb2, serial ));
	 serial++ )  {
	if ( !strcasecmp( (*parament)[nodeindex].string, node ) )  {
	    printf( "%s\t%s\n", (*parament)[0].string, node );
	}
    }
}
